/*
 *   Copyright (C) 2009-2015 by Jonathan Naylor G4KLX
 *
 *   This program is free software; you can redistribute it and/or modify
 *   it under the terms of the GNU General Public License as published by
 *   the Free Software Foundation; either version 2 of the License, or
 *   (at your option) any later version.
 *
 *   This program is distributed in the hope that it will be useful,
 *   but WITHOUT ANY WARRANTY; without even the implied warranty of
 *   MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
 *   GNU General Public License for more details.
 *
 *   You should have received a copy of the GNU General Public License
 *   along with this program; if not, write to the Free Software
 *   Foundation, Inc., 675 Mass Ave, Cambridge, MA 02139, USA.
 */

#include "Config.h"
#include "Globals.h"
#include "DMRRX.h"

const q15_t THRESHOLD = 3125;

const uint32_t PLLMAX = 0x10000U;
const uint32_t PLLINC = PLLMAX / DMR_RADIO_SYMBOL_LENGTH;
const uint32_t INC    = PLLINC / 32U;

const uint8_t DATA_SYNC_ERRS = 4U;

const unsigned int MAX_SYNC_BITS = 13U * DMR_FRAME_LENGTH_BITS;

const uint8_t BIT_MASK_TABLE[] = {0x80U, 0x40U, 0x20U, 0x10U, 0x08U, 0x04U, 0x02U, 0x01U};

#define WRITE_BIT(p,i,b) p[(i)/8] = (b) ? (p[(i)/8] | BIT_MASK_TABLE[(i)%8]) : (p[(i)/8] & ~BIT_MASK_TABLE[(i)%8])
#define READ_BIT(p,i)    (p[(i)/8] & BIT_MASK_TABLE[(i)%8])

CDMRRX::CDMRRX() :
m_pll(),
m_prev(),
m_patternBuffer(0U),
m_buffer(),
m_dataPtr(0U),
m_endPtr(),
m_index(0U),
m_control(0x00U),
m_state(DMR_NONE),
m_dataBits()
{
  m_pll[0U] = 0U;
  m_pll[1U] = 0U;

  m_prev[0U] = false;
  m_prev[1U] = false;

  m_endPtr[0U] = 999U;
  m_endPtr[1U] = 999U;

  m_dataBits[0U] = 0U;
  m_dataBits[1U] = 0U;
}

void CDMRRX::samples(const q15_t* samples, const uint8_t* control, uint8_t length)
{
  for (uint16_t i = 0U; i < length; i++) {
    switch (control[i]) {
      case MARK_SLOT1:
        m_dataPtr       = 0U;
        m_state         = DMR_SLOT1;
        m_index         = 0U;
        m_control       = 0x00U;
        m_patternBuffer = 0x00U;
        break;
      case MARK_SLOT2:
        m_dataPtr       = 0U;
        m_state         = DMR_SLOT2;
        m_index         = 1U;
        m_control       = 0x80U;
        m_patternBuffer = 0x00U;
        break;
      case MARK_END:
        m_state         = DMR_NONE;
        break;
      default:
        break;
    }

    if (m_state == DMR_NONE)
      continue;

    bool bit = samples[i] < 0;

    if (bit != m_prev[m_index]) {
      if (m_pll[m_index] < (PLLMAX / 2U))
        m_pll[m_index] += INC;
      else
        m_pll[m_index] -= INC;
    }

    m_prev[m_index] = bit;

    m_pll[m_index] += PLLINC;

    if (m_pll[m_index] >= PLLMAX) {
      m_pll[m_index] -= PLLMAX;

      if (samples[i] < -THRESHOLD) {
        processBit(false);
        processBit(true);
      } else if (samples[i] < 0) {
        processBit(false);
        processBit(false);
      } else if (samples[i] < THRESHOLD) {
        processBit(true);
        processBit(false);
      } else {
        processBit(true);
        processBit(true);
      }
    }
  }
}

void CDMRRX::processBit(bool b)
{
  m_patternBuffer <<= 1;
  if (b)
    m_patternBuffer |= 0x01U;
    
  WRITE_BIT(m_buffer, m_dataPtr, b);

  if (m_dataPtr == m_endPtr[m_index]) {
    uint8_t frame[DMR_FRAME_LENGTH_BYTES + 1U];
    uint8_t* data = frame + 1U;
    frame[0U] = m_control;

    unsigned int ptr = m_endPtr[m_index] - DMR_FRAME_LENGTH_BITS - 1U;
    for (unsigned int i = 0U; i < DMR_FRAME_LENGTH_BITS; i++, ptr++) {
      bool b = READ_BIT(m_buffer, ptr);
      WRITE_BIT(data, i, b);
    }

    serial.writeDMRData(frame, DMR_FRAME_LENGTH_BYTES + 1U);
  }

  if (countBits((m_patternBuffer & DMR_SYNC_BITS_MASK) ^ DMR_DATA_SYNC_BITS) <= DATA_SYNC_ERRS) {
    if (m_endPtr[m_index] == 999U) {
      io.setDecode(true);
    } else {
      io.setDecode(false);
      m_endPtr[m_index] = 999U;
    }

    m_endPtr[m_index] = m_dataPtr + 108U;
    m_dataBits[m_index] = 0U;
    m_control |= 0x40U;
  }

  if (countBits((m_patternBuffer & DMR_SYNC_BITS_MASK) ^ DMR_VOICE_SYNC_BITS) <= DATA_SYNC_ERRS) {
    if (m_endPtr[m_index] == 999U)
      io.setDecode(true);
    m_endPtr[m_index] = m_dataPtr + 108U;
    m_dataBits[m_index] = 0U;
    m_control |= 0x20U;
  }

  m_dataPtr++;

  if (m_endPtr[m_index] != 999U) {
    m_dataBits[m_index]++;

    if (m_dataBits[m_index] >= MAX_SYNC_BITS) {
      DEBUG1("DMRRX: Sync timed out, lost lock");
      serial.writeDMRLost(m_control);
      m_endPtr[m_index] = 999U;
      io.setDecode(false);
    }
  }
}

uint8_t CDMRRX::countBits(uint64_t pattern) const
{
  uint8_t count;
  
  for (count = 0U; pattern; count++)
    pattern &= pattern - 1U;

  return count;
}

