/*
 *   Copyright (C) 2015 by Jonathan Naylor G4KLX
 *
 *   This program is free software; you can redistribute it and/or modify
 *   it under the terms of the GNU General Public License as published by
 *   the Free Software Foundation; either version 2 of the License, or
 *   (at your option) any later version.
 *
 *   This program is distributed in the hope that it will be useful,
 *   but WITHOUT ANY WARRANTY; without even the implied warranty of
 *   MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
 *   GNU General Public License for more details.
 *
 *   You should have received a copy of the GNU General Public License
 *   along with this program; if not, write to the Free Software
 *   Foundation, Inc., 675 Mass Ave, Cambridge, MA 02139, USA.
 */

#include <Arduino.h>

#include "Config.h"
#include "Globals.h"
#include "IO.h"

// Generated using rcosdesign(0.2, 8, 5, 'sqrt') in MATLAB
static q15_t   C4FSK_FILTER[] = {401, 104, -340, -731, -847, -553, 112, 909, 1472, 1450, 683, -675, -2144, -3040, -2706, -770, 2667, 6995,
                                 11237, 14331, 15464, 14331, 11237, 6995, 2667, -770, -2706, -3040, -2144, -675, 683, 1450, 1472, 909, 112,
                                 -553, -847, -731, -340, 104, 401, 0};
const uint16_t C4FSK_FILTER_LEN = 42U;

// Generated using gaussfir(0.5, 4, 5) in MATLAB
static q15_t   GMSK_FILTER[] = {8, 104, 760, 3158, 7421, 9866, 7421, 3158, 760, 104, 8, 0};
const uint16_t GMSK_FILTER_LEN = 12U;

const uint16_t DC_OFFSET = 2048U;

const int PIN_COS    = 7;
const int PIN_PTT    = 8;
const int PIN_COSLED = 11;

extern "C" {
  void ADC_Handler()
  {
    if (ADC->ADC_ISR & ADC_ISR_EOC7)          // Ensure there was an End-of-Conversion and we read the ISR reg
      io.interrupt();
  }
}

CIO::CIO() :
m_started(false),
m_rxBuffer(),
m_txBuffer(),
m_C4FSKFilter(),
m_GMSKFilter(),
m_C4FSKState(),
m_GMSKState(),
m_pttInvert(false),
m_rxLevel(128 * 128),
m_txLevel(128 * 128),
m_ledCount(0U),
m_ledValue(true),
m_overflow(false)
{
  ::memset(m_C4FSKState, 0x00U, 70U * sizeof(q15_t));
  ::memset(m_GMSKState,  0x00U, 40U * sizeof(q15_t));

  m_C4FSKFilter.numTaps = C4FSK_FILTER_LEN;
  m_C4FSKFilter.pState  = m_C4FSKState;
  m_C4FSKFilter.pCoeffs = C4FSK_FILTER;

  m_GMSKFilter.numTaps = GMSK_FILTER_LEN;
  m_GMSKFilter.pState  = m_GMSKState;
  m_GMSKFilter.pCoeffs = GMSK_FILTER;

  // Set up the TX, COS and LED pins
  pinMode(PIN_PTT,    OUTPUT);
  pinMode(PIN_COSLED, OUTPUT);
  pinMode(PIN_LED,    OUTPUT);
}

void CIO::start()
{
  if (m_started)
    return;

  // Set up the ADC
  NVIC_EnableIRQ(ADC_IRQn);                   // Enable ADC interrupt vector
  ADC->ADC_IDR  = 0xFFFFFFFF;                 // Disable interrupts
  ADC->ADC_IER  = 0x80;                       // Enable AD7 End-Of-Conv interrupt (Arduino pin A0)
  ADC->ADC_CHDR = 0xFFFF;                     // Disable all channels
  ADC->ADC_CHER = 0x80;                       // Enable just A0
  ADC->ADC_CGR  = 0x15555555;                 // All gains set to x1
  ADC->ADC_COR  = 0x00000000;                 // All offsets off
  ADC->ADC_MR   = (ADC->ADC_MR & 0xFFFFFFF0) | (1 << 1) | ADC_MR_TRGEN;  // 1 = trig source TIO from TC0

  // Set up the timer
  pmc_enable_periph_clk(TC_INTERFACE_ID + 0*3+0) ;  // Clock the TC0 channel 0
  TcChannel* t = &(TC0->TC_CHANNEL)[0];       // Pointer to TC0 registers for its channel 0
  t->TC_CCR = TC_CCR_CLKDIS;                  // Disable internal clocking while setup regs
  t->TC_IDR = 0xFFFFFFFF;                     // Disable interrupts
  t->TC_SR;                                   // Read int status reg to clear pending
  t->TC_CMR = TC_CMR_TCCLKS_TIMER_CLOCK1 |    // Use TCLK1 (prescale by 2, = 42MHz)
              TC_CMR_WAVE |                   // Waveform mode
              TC_CMR_WAVSEL_UP_RC |           // Count-up PWM using RC as threshold
              TC_CMR_EEVT_XC0 |               // Set external events from XC0 (this setup TIOB as output)
              TC_CMR_ACPA_CLEAR | TC_CMR_ACPC_CLEAR |
              TC_CMR_BCPB_CLEAR | TC_CMR_BCPC_CLEAR;
  t->TC_RC  = 1750;                           // Counter resets on RC, so sets period in terms of 42MHz clock (was 875)
  t->TC_RA  = 880;                            // Roughly square wave (was 440)
  t->TC_CMR = (t->TC_CMR & 0xFFF0FFFF) | TC_CMR_ACPA_CLEAR | TC_CMR_ACPC_SET;  // Set clear and set from RA and RC compares
  t->TC_CCR = TC_CCR_CLKEN | TC_CCR_SWTRG ;  // re-enable local clocking and switch to hardware trigger source.

  // Set up the DAC
  pmc_enable_periph_clk(DACC_INTERFACE_ID);   // Start clocking DAC
  DACC->DACC_CR = DACC_CR_SWRST;              // Reset DAC
  DACC->DACC_MR =
    DACC_MR_TRGEN_EN | DACC_MR_TRGSEL(1) |    // Trigger 1 = TIO output of TC0
    (0 << DACC_MR_USER_SEL_Pos) |             // Select channel 0
    DACC_MR_REFRESH(0x0F) |                   // Bit of a guess... I'm assuming refresh not needed at 24kHz
    (24 << DACC_MR_STARTUP_Pos);              // 24 = 1536 cycles which I think is in range 23..45us since DAC clock = 42MHz
  DACC->DACC_IDR  = 0xFFFFFFFF;               // No interrupts
  DACC->DACC_CHER = DACC_CHER_CH0 << 0;       // Enable chan0

  digitalWrite(PIN_PTT, m_pttInvert ? HIGH : LOW);
  digitalWrite(PIN_COSLED, LOW);
  digitalWrite(PIN_LED, HIGH);

  m_started = true;
}

void CIO::process()
{
  m_ledCount++;
  if (m_started) {
    if (m_ledCount >= 24000U) {
      m_ledCount = 0U;
      m_ledValue = !m_ledValue;
      digitalWrite(PIN_LED, m_ledValue ? HIGH : LOW);
    }
  } else {
    if (m_ledCount >= 240000U) {
      m_ledCount = 0U;
      m_ledValue = !m_ledValue;
      digitalWrite(PIN_LED, m_ledValue ? HIGH : LOW);
    }
    return;
  }

  // Switch off the transmitter if needed
  if (m_txBuffer.getData() == 0U && tx) {
    tx = false;
    digitalWrite(PIN_PTT, m_pttInvert ? HIGH : LOW);
  }

  if (m_rxBuffer.getData() >= RX_BLOCK_SIZE) {
    q15_t   samples[RX_BLOCK_SIZE];
    uint8_t control[RX_BLOCK_SIZE];

    for (uint16_t i = 0U; i < RX_BLOCK_SIZE; i++) {
      uint16_t sample;
      m_rxBuffer.get(sample, control[i]);

      // Detect ADC overflow
      if (sample <= 1U || sample >= 4094U)
        m_overflow = true;

      q15_t res1 = q15_t(sample) - DC_OFFSET;
      q31_t res2 = res1 * m_rxLevel;
      samples[i] = q15_t(__SSAT((res2 >> 15), 16));
    }

    if (modemState == STATE_IDLE) {
      if (dstarEnable) {
        q15_t GMSKVals[RX_BLOCK_SIZE];
        ::arm_fir_fast_q15(&m_GMSKFilter, samples, GMSKVals, RX_BLOCK_SIZE);

        dstarRX.samples(GMSKVals, RX_BLOCK_SIZE);
      }

      if (dmrEnable || ysfEnable) {
        q15_t C4FSKVals[RX_BLOCK_SIZE];
        ::arm_fir_fast_q15(&m_C4FSKFilter, samples, C4FSKVals, RX_BLOCK_SIZE);

        if (dmrEnable)
          dmrIdleRX.samples(C4FSKVals, RX_BLOCK_SIZE);

        if (ysfEnable)
          ysfRX.samples(C4FSKVals, RX_BLOCK_SIZE);
      }
    } else if (modemState == STATE_DSTAR) {
      if (dstarEnable) {
        q15_t GMSKVals[RX_BLOCK_SIZE];
        ::arm_fir_fast_q15(&m_GMSKFilter, samples, GMSKVals, RX_BLOCK_SIZE);

        dstarRX.samples(GMSKVals, RX_BLOCK_SIZE);
      }
    } else if (modemState == STATE_DMR) {
      if (dmrEnable) {
        q15_t C4FSKVals[RX_BLOCK_SIZE];
        ::arm_fir_fast_q15(&m_C4FSKFilter, samples, C4FSKVals, RX_BLOCK_SIZE);

        // If the transmitter isn't on, use the DMR idle RX to detect the wakeup CSBKs
        if (tx)
          dmrRX.samples(C4FSKVals, control, RX_BLOCK_SIZE);
        else
          dmrIdleRX.samples(C4FSKVals, RX_BLOCK_SIZE);
      }
    } else if (modemState == STATE_YSF) {
      if (ysfEnable) {
        q15_t C4FSKVals[RX_BLOCK_SIZE];
        ::arm_fir_fast_q15(&m_C4FSKFilter, samples, C4FSKVals, RX_BLOCK_SIZE);

        ysfRX.samples(C4FSKVals, RX_BLOCK_SIZE);
      }
    } else if (modemState == STATE_CALIBRATE) {
      q15_t GMSKVals[RX_BLOCK_SIZE];
      ::arm_fir_fast_q15(&m_GMSKFilter, samples, GMSKVals, RX_BLOCK_SIZE);

      calRX.samples(GMSKVals, RX_BLOCK_SIZE);
    }
  }
}

void CIO::write(q15_t* samples, uint16_t length, const uint8_t* control)
{
  if (!m_started)
    return;

  // Switch the transmitter on if needed
  if (!tx) {
    tx = true;
    digitalWrite(PIN_PTT, m_pttInvert ? LOW : HIGH);
  }

  for (uint16_t i = 0U; i < length; i++) {
    q31_t res1 = samples[i] * m_txLevel;
    q15_t res2 = q15_t(__SSAT((res1 >> 15), 16));
    if (control == NULL)
      m_txBuffer.put(uint16_t(res2 + DC_OFFSET), MARK_NONE);
    else
      m_txBuffer.put(uint16_t(res2 + DC_OFFSET), control[i]);
  }
}

uint16_t CIO::getSpace() const
{
  return m_txBuffer.getSpace();
}

void CIO::interrupt()
{
  uint8_t control = MARK_NONE;
  uint16_t sample = DC_OFFSET;

  if (m_txBuffer.getData() > 0U)
    m_txBuffer.get(sample, control);
    
  DACC->DACC_CDR = sample;

  sample = ADC->ADC_CDR[7];

  m_rxBuffer.put(sample, control);
}

void CIO::setDecode(bool dcd)
{
  digitalWrite(PIN_COSLED, dcd ? HIGH : LOW);
}

void CIO::setParameters(bool rxInvert, bool txInvert, bool pttInvert, uint8_t rxLevel, uint8_t txLevel)
{
  m_pttInvert = pttInvert;

  m_rxLevel = q15_t(rxLevel * 128);
  m_txLevel = q15_t(txLevel * 128);
  
  if (rxInvert)
    m_rxLevel = -m_rxLevel;
  
  if (txInvert)
    m_txLevel = -m_txLevel;
}

bool CIO::hasADCOverflow()
{
  bool overflow = m_overflow;
  
  m_overflow = false;
  
  return overflow;
}

bool CIO::hasBufferOverflow()
{
  return m_rxBuffer.hasOverflow();
}

